#define _XOPEN_SOURCE	500

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "cf_chunk_file.h"
#include "cf_log.h"
#include "cf_version.h"

Cf_chunk_file cf_chunk_file_alloc() {
	Cf_chunk_file ret;

	ret = (Cf_chunk_file) malloc(sizeof(Cf_chunk_file_base));
	if (!ret)
		printf("Could not alloc in %s\n", __FUNCTION__);

	cf_chunk_file_set_header(ret, cf_chunk_header_alloc());
	cf_chunk_file_set_toc(ret, cf_chunk_toc_alloc());
	cf_chunk_file_set_old_toc(ret, cf_chunk_toc_alloc());

	cf_chunk_file_set_buffer(ret, cf_chunk_buffer_alloc());

	pthread_rwlock_init(&cf_chunk_file_get_lock(ret), NULL);

	return ret;
}

void cf_chunk_file_dealloc(Cf_chunk_file f) {
	cf_chunk_header_dealloc(cf_chunk_file_get_header(f));
	cf_chunk_toc_dealloc(cf_chunk_file_get_toc(f));
	cf_chunk_toc_dealloc(cf_chunk_file_get_old_toc(f));
	cf_chunk_buffer_dealloc(cf_chunk_file_get_buffer(f));

	/* Probably should check if still locked, would indicate a bug */
	pthread_rwlock_destroy(&cf_chunk_file_get_lock(f));

	free(f);
}


Cf_chunk_file cf_chunk_open(int fd, Cf_config c, int compress) {
	Cf_chunk_file ret;
	Cf_chunk_header h;
	Cf_chunk_toc t;		/* Working toc */
	Cf_chunk_toc ot;	/* Old toc */
	int res;

	ret = cf_chunk_file_alloc();

	h = cf_chunk_file_get_header(ret);
	t = cf_chunk_file_get_toc(ret);
	ot = cf_chunk_file_get_old_toc(ret);

	cf_chunk_file_set_dirty(ret, 0);
	cf_chunk_file_set_fd(ret, fd);
	cf_chunk_file_set_config(ret, c);

	/* Read header */
	res = pread(fd, h, sizeof(Cf_chunk_header_base), 0);

	if (res != sizeof(Cf_chunk_header_base)) {
		/* This is a NEW file */ 

		/* Set defaults in header */
		cf_chunk_header_set_data_size(h, 0);
		cf_chunk_header_set_toc_size(h, 0);
		cf_chunk_header_set_toc_offset(h, 0);
		cf_chunk_header_set_chunk_size(h, cf_config_get_chunk_size(c));

		if (compress)
			cf_chunk_header_set_coding(h, cf_config_get_compression_id(c));
		else
			cf_chunk_header_set_coding(h, CF_RAW);

		cf_chunk_header_set_version(h, CF_VERSION);

		return ret;
	}

	/* Behind the line the TOC is read from disk */

	/* Resize the toc */
	cf_chunk_toc_resize(t, cf_chunk_header_get_toc_size(h));
	cf_chunk_toc_resize(ot, cf_chunk_header_get_toc_size(h));

	/* Read the toc */
	pread(fd, cf_chunk_toc_get_entries(t), sizeof(Cf_toc_entry) * cf_chunk_header_get_toc_size(h),
		cf_chunk_header_get_toc_offset(h));
	/* Read the toc (again) */
	pread(fd, cf_chunk_toc_get_entries(ot), sizeof(Cf_toc_entry) * cf_chunk_header_get_toc_size(h),
		cf_chunk_header_get_toc_offset(h));


	return ret;
}

void cf_chunk_close(Cf_chunk_file f) {
	cf_chunk_file_dealloc(f);
}

int cf_chunk_size(Cf_chunk_file f) {
	Cf_chunk_header h = cf_chunk_file_get_header(f);

	return cf_chunk_header_get_data_size(h);
}

/*
	Return chunk data (read from disk if necessary)
*/
/*
Cf_chunk_data cf_chunk_get_chunk(Cf_chunk_file f, int idx) {
	Cf_chunk_buffer b;

	b = cf_chunk_file_get_buffer(f);

	return cf_chunk_buffer_get_buffer(b, idx);
}
*/

/*
	Return chunk data (read from disk if necessary)
*/
void cf_chunk_preload_chunks(Cf_chunk_file f, int from, int to) {
	Cf_chunk_data ret;
	Cf_chunk_buffer b;
	int i;

	b = cf_chunk_file_get_buffer(f);

	/* Add 1 */
	to++;

	if (to > cf_chunk_buffer_get_buf_sz(b)) {
		/* Resize chunk buffer */
		cf_chunk_buffer_resize(b, to);
	}

	for (i=from; i<to; i++) {
		if (cf_chunk_buffer_get_buffer(b, i) == NULL) {
			/* Chunk not in buffer */

			/* cf_chunk_read() will alloc a new chunk or read one from disk */
			ret = cf_chunk_read(f, i);

			cf_chunk_buffer_add(b, i, ret);
		}
	}

	return;
}

void cf_chunk_unload_chunks(Cf_chunk_file f) {
	Cf_chunk_buffer b = cf_chunk_file_get_buffer(f);

	/* Check if there are too many buffers */
	if (cf_chunk_buffer_get_buf_used(b) > cf_config_get_chunk_max(cf_chunk_file_get_config(f))) {
		/* Throw / write all the buffer away */
		cf_chunk_flush(f);
	}

}

Cf_chunk_data cf_chunk_read(Cf_chunk_file f, int idx) {
	Cf_chunk_toc 		t = cf_chunk_file_get_toc(f);
	Cf_chunk_header 	h = cf_chunk_file_get_header(f);
	Cf_config 		c = cf_chunk_file_get_config(f);
	Cf_decompressor_f 	decomp = cf_config_get_decompressor(c);
	Cf_error_f		err = cf_config_get_error(c);
	Cf_chunk_data d;
	Cf_chunk_data raw;
	int fd;
	int offset;
	int sz;
	int osz;
	int res;

	osz = cf_chunk_header_get_chunk_size(h);

	/* Create a new data chunk */
	d = cf_chunk_data_alloc(osz);

	/* 
		Allocate a data chunk

		If present in toc read:
		1) read from disk
		2) decompress
		3) set the end of chunk

		If not present:
		1) extend toc if necessary
	*/

	if (idx >= cf_chunk_toc_get_size(t)) {
		/* Extend the toc */
		cf_chunk_toc_resize(t, idx+1);

		/* Mark as empty entry */
		cf_chunk_toc_set_entry_offset(t, idx, -1);
		cf_chunk_toc_set_entry_size(t, idx, 0);
	} else {
		/* Read from disk into data chunk */
		/* Get size and offset in the file */
		offset = cf_chunk_toc_get_entry_offset(t, idx);
		sz = cf_chunk_toc_get_entry_size(t, idx);

		/* Exist in toc but no data: return empty buffer */
		if (sz == 0)
			return d;
#ifdef  CF_DEBUG
if (abs(sz)>cf_chunk_header_get_chunk_size(h))
	cf_printf("sze>chunk size: |%d|>%d\n", sz, cf_chunk_header_get_chunk_size(h));
#endif

#ifdef CF_DEBUG
			cf_printf("rd @%d size %d\n", offset, sz);
#endif

		/* Get fd */
		fd = cf_chunk_file_get_fd(f);

		if (sz > 0) {
			/* Get mem for data */
			raw = cf_chunk_data_alloc(cf_chunk_header_get_chunk_size(h));

			/* Read from disk */
			/*
			res = pread(fd, cf_chunk_data_get_data(raw), sz, offset);
			*/
			res = cf_chunk_read_raw(fd, t, idx, raw);

			if (res != sz) 
				cf_printf("%d in %s: pread() gave %d instead of %d\n", __LINE__, __FUNCTION__, res, sz);

			/* Decompress into data chunk d */
			res = decomp(cf_chunk_data_get_data(d), &osz, cf_chunk_data_get_data(raw), sz);

			/* Problem display message */
			if (res != 0) {
				err(res);
#ifdef CF_DEBUG
				exit(1);
#endif
			}

			/* Set the end position */
			cf_chunk_data_set_end(d, osz);

			/* Throw away tmp 'raw' chunk */
			cf_chunk_data_dealloc(raw);
		} else {
			/* sz<0 -> chunk is NOT compressed */
			/* Read from disk */
			/*
			res = pread(fd, cf_chunk_data_get_data(d), abs(sz), offset);
			*/
			res = cf_chunk_read_raw(fd, t, idx, d);

			if (res != abs(sz)) 
				cf_printf("%d in %s: pread() gave %d instead of %d\n", __LINE__, __FUNCTION__, res, abs(sz));

			/* Set the end position */
			cf_chunk_data_set_end(d, abs(sz));
		}
	}

	return d;	
}

int cf_chunk_write_raw(int fd, Cf_chunk_toc t, int idx, Cf_chunk_data d) {
	return pwrite(fd, cf_chunk_data_get_data(d), abs(cf_chunk_toc_get_entry_size(t, idx)), cf_chunk_toc_get_entry_offset(t,idx));
}

int cf_chunk_read_raw(int fd, Cf_chunk_toc t, int idx, Cf_chunk_data d) {
	return pread(fd, cf_chunk_data_get_data(d), abs(cf_chunk_toc_get_entry_size(t, idx)), cf_chunk_toc_get_entry_offset(t,idx));
}

#ifdef CF_DEBUG_L3
void cf_chunk_file_traverse(Cf_chunk_file f) {
	Cf_chunk_toc 		t = cf_chunk_file_get_toc(f);
	Cf_chunk_data tmp, out;
	Cf_chunk_header 	h = cf_chunk_file_get_header(f);
	Cf_config 		c = cf_chunk_file_get_config(f);
	Cf_decompressor_f 	decomp = cf_config_get_decompressor(c);
	Cf_error_f		err = cf_config_get_error(c);
	int osz = cf_chunk_header_get_chunk_size(h);
	int sz;
	int i;
	int fd = cf_chunk_file_get_fd(f);
	int res;
	int loaded;

	off_t old = lseek(fd, 0, SEEK_CUR);

	sz = cf_chunk_toc_get_size(t);

	tmp = cf_chunk_data_alloc(osz);
	out = cf_chunk_data_alloc(osz);

	for (i=0; i<sz; i++) {
		loaded = cf_chunk_read_raw(fd, t, i, tmp);

		osz = cf_chunk_header_get_chunk_size(h);

		res = cf_chunk_toc_get_entry_size(t,i);

		if (loaded != res)
			printf("Read %d instead of %d\n", loaded, res);

		if (res > 0) {
			res = decomp(cf_chunk_data_get_data(out), &osz, 
				cf_chunk_data_get_data(tmp), cf_chunk_toc_get_entry_size(t,i));
		} else {
			/* Bit bogus but basically there was no decompression problem */
			res = 0;
		} 

		/* Problem display message */
		if (res != 0) {
			cf_printf("Reading chunk %d\n", i);
			err(res);
			cf_chunk_toc_print(t);
			exit(1);
		}
	}

	cf_chunk_data_dealloc(tmp);
	cf_chunk_data_dealloc(out);

	/* Restore fd position */
	lseek(fd, old, SEEK_SET);

}
#endif


/*
	This function write all the chunk that are currently
	in the 'buffer' of the give file to disk.

	3 steps

	1) the chunks in the buffer are compressed

	2) a user specified plugin function will be called. 
	This plugin will write the chunks to disk in any order 
	it sees fit. As long as the toc and header are correct 
	(and no data is lost)

	3) the chunk buffer is emptied (chunks are freed)

	Changes: toc, header
*/	

void cf_chunk_flush(Cf_chunk_file f) {
	Cf_chunk_toc 		t = cf_chunk_file_get_toc(f);
	Cf_chunk_toc 		ot = cf_chunk_file_get_old_toc(f);
	Cf_chunk_header 	h = cf_chunk_file_get_header(f);
	Cf_config 		c = cf_chunk_file_get_config(f);
	Cf_compressor_f 	comp = cf_config_get_compressor(c);
	Cf_error_f		err = cf_config_get_error(c);
	Cf_buffersizer_f	sizer = cf_config_get_buffersizer(c);
	Cf_writer_f		writer = cf_config_get_writer(c);
	Cf_chunk_data d;
	Cf_chunk_data raw;
	Cf_chunk_data tmp;
	Cf_chunk_buffer b;
	int sz;
	int osz;
	int i;
	int chunk_size = cf_chunk_header_get_chunk_size(h);
	int res;


	b = cf_chunk_file_get_buffer(f);
	sz = cf_chunk_buffer_get_buf_sz(b);

#ifdef CF_DEBUG_L2
	cf_chunk_toc_print(t);
	cf_chunk_buffer_print(b);
#endif

	/* Set the toc size in the header */
	cf_chunk_header_set_toc_size(h, cf_chunk_toc_get_size(t));

	if (cf_chunk_file_get_dirty(f)) {
		/* Compress all the buffers in the buffer */
		for (i=0; i<sz; i++) {
			d = cf_chunk_buffer_get_buffer(b, i);

			if (!d)
				continue;


			if (cf_chunk_data_get_dirty(d)) {
				if (cf_chunk_data_get_end(d) ) {
					/* Copy d pointer */
					tmp = d;

					/* Create destination chunk */
					osz = sizer(cf_chunk_header_get_chunk_size(h));
					raw = cf_chunk_data_alloc(osz);

					/* Compress into data chunk raw */
					res =comp(cf_chunk_data_get_data(raw), &osz,
					cf_chunk_data_get_data(d), cf_chunk_data_get_end(d));

					/* Problem display message */
					if (res != 0)
						err(res);

					/* If data is uncompressible we just skip */
					if (osz < cf_chunk_data_get_end(d)) {
						/* Set the compressed size */	
						cf_chunk_data_set_end(raw, osz);
						cf_chunk_data_set_dirty(raw, 1);

						/* Replace 'd' by raw */
						cf_chunk_buffer_set_buffer(b, i, raw);
#ifdef CF_DEBUG_L4
					{
					Cf_decompressor_f 	decomp = cf_config_get_decompressor(c);
					int tmp_sz = chunk_size;

					printf("Decompressing chunk %d\n", i);
					res =decomp(cf_chunk_data_get_data(d), &tmp_sz, cf_chunk_data_get_data(raw), osz);
					printf("Decompressing chunk %d done\n", i);

					if (res != 0)
						err(res);
					}
#endif

						/* Dealloc old 'd' */
						cf_chunk_data_dealloc(d);

						/* Update the size in the toc */
						cf_chunk_toc_set_entry_size(t, i, osz);
					} else {
#ifdef CF_DEBUG_L3
						printf("No compressing chunk\n");
#endif
						cf_chunk_data_dealloc(raw);
						/* Update the size in the toc: NEGATIVE!!! */
						cf_chunk_toc_set_entry_size(t, i, -cf_chunk_data_get_end(d));
					}
				} else {
					cf_chunk_toc_set_entry_size(t, i, 0);
				}
#ifdef CF_DEBUG_L3
					printf("Chunk %d raw %d %d %d %d\n"
,i
, (unsigned char) cf_chunk_buffer_get_buffer(b,i)->data[0]
, (unsigned char) cf_chunk_buffer_get_buffer(b,i)->data[1]
, (unsigned char) cf_chunk_buffer_get_buffer(b,i)->data[2]
, (unsigned char) cf_chunk_buffer_get_buffer(b,i)->data[3]
);
#endif
			} else {
				cf_chunk_data_dealloc(d);
				cf_chunk_buffer_del(b, i);
			}
		}

#ifdef CF_DEBUG_L2
	cf_chunk_buffer_print(b);
#endif



		/* Call the plugin code */
		writer( cf_chunk_file_get_fd(f), chunk_size, h, b, t, ot);

	        /* The first flush() on a brand new file has no old toc... we fake it */
	        if (cf_chunk_toc_get_size(ot) < cf_chunk_toc_get_size(t)) {
	                cf_chunk_toc_resize(ot, cf_chunk_toc_get_size(t));
	        }

		/* Toc got out of sync -> copy */
	    	cf_chunk_toc_copy(t, ot);


#ifdef CF_DEBUG_L2
	cf_chunk_buffer_print(b);
#endif
	}

#ifdef CF_DEBUG_L3
	cf_chunk_file_traverse(f);
#endif

	/* Deallocate all the chunk in the buffer */
	sz = cf_chunk_buffer_get_buf_sz(b);
	for (i=0; i<sz; i++) {
		d = cf_chunk_buffer_get_buffer(b, i);

		if (d) {
			cf_chunk_data_dealloc(d);
			cf_chunk_buffer_del(b, i);
		}
	}

	/* Resize the buffer */
	cf_chunk_buffer_resize(b, 0);
	cf_chunk_buffer_set_buf_used(b, 0);

	cf_chunk_file_set_dirty(f, 0);

#ifdef CF_DEBUG_L2
	cf_chunk_toc_print(t);
#endif

}


void cf_chunk_write_header(Cf_chunk_file f) {
	int fd = cf_chunk_file_get_fd(f);
	Cf_chunk_header h = cf_chunk_file_get_header(f);


	pwrite(fd, h, sizeof(Cf_chunk_header_base), 0);
}
