/*
#include <config.h>
*/
#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#define FUSE_USE_VERSION 25

#include <fuse.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>
#include <string.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif

#include <pthread.h>

#include "cf_chunk_file.h"
#include "cf_log.h"
#include "cf_file.h"
#include "cf_scan.h"
#include "cf_info.h"
#include "cf_map.h"
#include "cf_config_file.h"
#include "cf_config.h"
#include "cf_version.h"

#define CF_PATH_MAX	4096

#define CF_DEBUG_OFF
#define CF_DEBUG_L2_OFF


/* GLOBALS */
Cf_config_file config_file;
Cf_config config;
Cf_info info;
Cf_map map;
pthread_mutex_t info_lock;

enum { start_, middle_, stop_, last_};

/* New function needed for FUSE 2.5.x */

static inline DIR *get_dirp(struct fuse_file_info *fi)
{
    return (DIR *) (uintptr_t) fi->fh;
}

static inline Cf_chunk_file get_chunk_file(struct fuse_file_info *fi)
{
    return (Cf_chunk_file) (uintptr_t) fi->fh;
}


/* ------------------------------------------------- */

static int cf_getattr(const char *path, struct stat *stbuf)
{
    int res;
    int fd;
    char xpath[CF_PATH_MAX];
    Cf_chunk_file f;

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = lstat(xpath, stbuf);
    if(res == -1)
        return -errno;

    if (S_ISDIR(stbuf->st_mode))
	return 0;


    if (S_ISREG(stbuf->st_mode) && (stbuf->st_size > 0)) {
        fd = open (xpath, O_RDONLY);
	if (fd != -1) {
            f = cf_chunk_open(fd, config, !cf_file_no_compress(config, path));
    	    stbuf->st_size = cf_chunk_size(f);

	    cf_chunk_close(f);
	    close(fd);
	} else {
	    cf_printf("File %s not found... in %s\n", xpath, __FUNCTION__);
	}
    }

    return 0;
}


static int cf_readlink(const char *path, char *buf, size_t size)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = readlink(xpath, buf, size - 1);
    if(res == -1)
        return -errno;

    buf[res] = '\0';
    return 0;
}

static int cf_opendir(const char *path, struct fuse_file_info *fi)
{
    DIR *dp;

    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    dp = opendir(xpath);
    if (dp == NULL)
        return -errno;

    fi->fh = (unsigned long) dp;
    return 0;
}

static int cf_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi)
{
    DIR *dp = get_dirp(fi);
    struct dirent *de;

    (void) path;
    seekdir(dp, offset);
    while ((de = readdir(dp)) != NULL) {
        struct stat st;
        memset(&st, 0, sizeof(st));
        st.st_ino = de->d_ino;
        st.st_mode = de->d_type << 12;
        if (filler(buf, de->d_name, &st, de->d_off))
            break;
    }

    return 0;
}

static int cf_releasedir(const char *path, struct fuse_file_info *fi)
{
    DIR *dp = get_dirp(fi);
    (void) path;
    closedir(dp);
    return 0;
}

static int cf_mknod(const char *path, mode_t mode, dev_t rdev)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = mknod(xpath, mode, rdev);
    if(res == -1) 
        return -errno;

    return 0;
}

static int cf_mkdir(const char *path, mode_t mode)
{
    int res;
    char xpath[CF_PATH_MAX];

    /* Some apps create too long paths ... */
    if (strlen(path) >= CF_PATH_MAX)
	return -ENAMETOOLONG;

    cf_path(config, path, xpath);

    res = mkdir(xpath, mode);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_unlink(const char *path)
{
    int res;
    char xpath[CF_PATH_MAX];
    Cf_chunk_file pos;
    Cf_chunk_header h;
    int fd;
    int size;

    cf_path(config, path, xpath);


    /* Create compFUSEd structure in that file */
    fd = open (xpath, O_RDONLY);
    pos = cf_chunk_open(fd, config, !cf_file_no_compress(config, path));
    h = cf_chunk_file_get_header(pos);

    size = cf_chunk_header_get_data_size(h);
    cf_chunk_close(pos);
    close(fd);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = unlink(xpath);
    if(res == -1)
        return -errno;


    /* Update total fs size */
    pthread_mutex_lock(&info_lock);
    cf_info_set_ucomp(info, cf_info_get_ucomp(info) - size);
    pthread_mutex_unlock(&info_lock);

    return 0;
}

static int cf_rmdir(const char *path)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = rmdir(xpath);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_symlink(const char *from, const char *to)
{
    int res;
    char xfrom[CF_PATH_MAX];
    char xto[CF_PATH_MAX];

    strcpy(xfrom, cf_config_get_backendpath(config));
    if (from[0] != '/')
	strcat(xfrom, "/");

    strcat(xfrom, from);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xfrom);
#endif 
    cf_path(config, to, xto);

    res = symlink(from, xto);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_rename(const char *from, const char *to)
{
    int res;
    char xfrom[CF_PATH_MAX];
    char xto[CF_PATH_MAX];

#ifdef CF_DEBUG_L2
cf_printf ("RENAME %s to %s\n", from, to);
#endif

    cf_path(config, from, xfrom);
    cf_path(config, to, xto);

    res = rename(xfrom, xto);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_link(const char *from, const char *to)
{
    int res;
    char xfrom[CF_PATH_MAX];
    char xto[CF_PATH_MAX];

    cf_path(config, from, xfrom);
    cf_path(config, to, xto);

    res = link(xfrom, xto);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_chmod(const char *path, mode_t mode)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = chmod(xpath, mode);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_chown(const char *path, uid_t uid, gid_t gid)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = lchown(xpath, uid, gid);
    if(res == -1)
        return -errno;

    return 0;
}

static int cf_truncate(const char *path, off_t size)
{
    int res;
    int fd;
    char xpath[CF_PATH_MAX];
    Cf_chunk_file f;
    Cf_chunk_header h;
    Cf_chunk_toc t;
    Cf_chunk_toc ot;
    Cf_chunk_data d;
    int offset = sizeof(Cf_chunk_header_base);
    int chk;
    off_t old_size;

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("TRUNCATE %s\n", xpath);
#endif 


    fd = open (xpath, O_RDWR);
    if (fd != -1) {
	f = cf_chunk_open(fd, config, !cf_file_no_compress(config, path));
	h = cf_chunk_file_get_header(f);
	t = cf_chunk_file_get_toc(f);
	ot = cf_chunk_file_get_old_toc(f);

	old_size = cf_chunk_header_get_data_size(h);

	pthread_mutex_lock(&info_lock);
    	cf_info_set_ucomp(info, cf_info_get_ucomp(info) - old_size + size);
	pthread_mutex_unlock(&info_lock);

	/* Update the total fs size */
    
	if (cf_chunk_header_get_coding(h) == CF_RAW) {
    	    /* This file is not compressed by us; there is not toc just a header! */

	    /* Adjust size in the header */
	    cf_chunk_header_set_data_size(h, size);

	    pwrite(fd, h, sizeof(Cf_chunk_header_base), 0);

	    /* Truncate the file */
	    res = ftruncate(fd, size+offset);
	} else {
	    /* Compressed file: we must adjust the toc and the header */

	    /* Adjust size in the header */
	    cf_chunk_header_set_data_size(h, size);

	    /* Find the chunk that needs to be shortened */
	    chk = size / cf_chunk_header_get_chunk_size(h);
	
    	    /* Preload the chunk */
	    cf_chunk_preload_chunks(f, chk, chk);
	    d = cf_chunk_get_chunk(f, chk);

	    cf_chunk_file_set_dirty(f, 1);
	    cf_chunk_data_set_dirty(d, 1);

	    /* Now reduce the size of the data in this chunk */
	    cf_chunk_data_set_end(d, size - chk*cf_chunk_header_get_chunk_size(h));

	    /* Dispose of chunk */
    	    cf_chunk_unload_chunks(f);

	    /* Reduce the toc size in the header */
	    cf_chunk_header_set_toc_size(h, chk+1);

	    /* 
	    pwrite(fd, h, sizeof(Cf_chunk_header_base), 0);
	    */

	    /* Now resize toc to retain only the require number of entries */
	    cf_chunk_toc_resize(t, chk+1);
	    cf_chunk_toc_resize(ot, chk+1);

	    /* Flush the file (it will be truncated by the chunk writer plugin */
	    cf_chunk_flush(f);
	}

	cf_chunk_close(f);
	close(fd);

	if(res == -1)
	    return -errno;

    }

    return 0;
}

static int cf_utime(const char *path, struct utimbuf *buf)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    res = utime(xpath, buf);
    if(res == -1)
        return -errno;

    return 0;
}


static int cf_open(const char *path, struct fuse_file_info *fi)
{
    Cf_chunk_file pos;
    char xpath[CF_PATH_MAX];
    int fd;
    Cf_map_entry e;

    /* Some apps create too long paths ... */
    if (strlen(path) >= CF_PATH_MAX)
	return -ENAMETOOLONG;

    cf_path(config, path, xpath);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, xpath);
#endif 

    pthread_mutex_lock(&cf_map_get_lock(map));

    /* Lookup in map */
    e = cf_map_findname(map, path);

    if (e) {
	/* Found */
	pos = cf_entry_get_file(e);
#ifdef CF_DEBUG
    cf_printf("%s opens %d (reopen)\n", __FUNCTION__, cf_entry_get_opens(e));
#endif 
    } else {    

#ifdef CF_DEBUG
    cf_printf("%s new open\n", __FUNCTION__);
#endif 
    	/* Open the file on the underlying fs */
    	/*
	IGNORE the flags! We always need to read the header even for a write only open..
    	*/
    	fd = open(xpath, O_RDWR);
    	if (fd == -1) {
    	    return -errno;
    	}

    	/* Create compFUSEd structure in that file */
    	pos = cf_chunk_open(fd, config, !cf_file_no_compress(config, path));
    }

    /* Add to map */
    cf_map_addname(map, path, pos);
	
    pthread_mutex_unlock(&cf_map_get_lock(map));

    fi->fh = (int) pos;

    return 0;
}

int sz2chk(int sz, int chk_sz) {
	int ret;

	ret = sz / chk_sz;
	if (sz % chk_sz) {
		ret++;
	}

	return ret;
}

static int cf_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
    int res = 0;
    Cf_chunk_file f = get_chunk_file(fi);
    Cf_chunk_header h = cf_chunk_file_get_header(f);
    Cf_chunk_data d;
    Cf_config c;
    int i;
    int len;
    int clen; /* copy length */
    int start, stop, last;
    int csz;
    int co;		/* Offset within chunk */
    int left;
    char *i_buffer;
    int i_size;
    int ops;
    (void) path;

#ifdef CF_DEBUG
cf_printf("READ %s\n", path);
#endif

    /* Normal read for files not compressed by us */
    if (cf_chunk_header_get_coding(h) == CF_RAW) {
	res = pread(cf_chunk_file_get_fd(f), buf, size, sizeof(Cf_chunk_header_base)+offset);

	if (res == -1)
		res = -errno;

	return res;
    }

    cf_chunk_file_rdlock(f);

    c = cf_chunk_file_get_config(f);

    /* Get actual chunk size from file rather than from config */
    csz = cf_chunk_header_get_chunk_size(h);

    /* Get number of last chunk */
    last = cf_chunk_toc_get_size(cf_chunk_file_get_toc(f)) - 1;
    
    if (offset > cf_chunk_header_get_data_size(h))
	return 0;

    if ((offset+size) > cf_chunk_header_get_data_size(h)) {
	#ifdef CF_DEBUG
   	cf_printf("%d beyond EOF (%d)\n", offset, cf_chunk_header_get_data_size(h));
	#endif

	size = cf_chunk_header_get_data_size(h) - offset;
    }

    /* Index of first chunk */
    start = offset / csz;
    /* Index of last chunk */
    stop = (offset+size-1) / csz;


    res = 0;
    left = size;

#ifdef CF_DEBUG
   cf_printf("start=%d stop=%d last=%d\n", start, stop, last);
#endif

    /* Preload all the chunks */
    cf_chunk_preload_chunks(f, start, stop);

    for (i=start;  i<stop+1; i++) {
	d = cf_chunk_get_chunk(f, i);
    	i_buffer = cf_chunk_data_get_data(d);
    	i_size = cf_chunk_data_get_end(d);

	len = csz;

	/* Chunk offset */
	co = 0;

	/* First chunk? */
	if (i == start) {
		/* [___<--->] */
		co = offset - i * csz;
	}

	/*
		co < i_size
		- chunk is full
		- chunk not full

		- first chunk  co > 0
		- last chunk	co = 0
		- other chunk   co = 0
	*/

	if (i == stop) {
		/* [<--->__] */
		/* copy from co (=0) to i_size */
		ops = stop_;
	} else {
		if (i == start) {
			/* [__<--->] */
			/* copy from co to csz (possibly fake) */
			ops = start_;
		} else {
			/* [<----->] */
			/* copy from co (=0) to csz (possibly fake) */
			ops = middle_;
		}
	}

	switch (ops) {
		case start_:
			/*
			co >= 0
			len = csz - co
			*/
			len = csz - co;

			if (co > i_size)
				clen = 0;
			else
				clen = i_size - co;

			break;

		case middle_:
			/*
			co = 0
			len = csz
			*/
			len = csz;
			clen = i_size;

			break;

		case stop_:
			/*
			co = 0
			len = csz - co
			*/
			if (i == last) {
				len = i_size;

				/* Do not read too much */
				if (len > left)
					len = left;

				clen = len;
			} else {
				len = csz;
				clen = i_size;

				/* Do not read too much */
				if (len > left)
					len = left;

				if (clen > left)
					clen = left;

			}

			break;
	};


#ifdef CF_DEBUG
			cf_printf("copy (II) start=%d stop=%d i=%d co=%d len=%d i_size=%d left=%d\n", start, stop, i, co, len, i_size, left);
#endif

	/* Copy to output */
	if (clen)
		memcpy(buf+res, i_buffer+co, clen);

	res += len;
	left -= len;
    }


    if (res > size)
	cf_printf("start=%d stop=%d res=%d size=%d offset=%d\n", start, stop, res, size, offset);

/*
#ifdef CF_DEBUG
	if (res != size) {
		cf_printf("ERROR: copying %d instead of %d in %s\n", res, size, __FUNCTION__);
	}
#endif
*/

#ifdef CF_DEBUG_L2
cf_printf("copying %d (%d) ", res, offset+size);
#endif


#ifdef CF_DEBUG_L2
cf_puts("copied");
#endif

    cf_chunk_unload_chunks(f);

    cf_chunk_file_unlock(f);


    return res;
}

static int cf_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    int res;
    Cf_chunk_file f = get_chunk_file(fi);
    Cf_chunk_header h;
    int i_size;
    char *i_buffer;
	int i;
	int csz;
	Cf_chunk_data d;
	int c_offset;
	int c_len;
	int written;
	int left;
	off_t old_size;
	off_t new_size = offset + size;
    int start, stop;

#ifdef CF_DEBUG
cf_printf("%s %s\n", __FUNCTION__, path);
#endif

	/* Get header */
	h = cf_chunk_file_get_header(f);

	old_size = cf_chunk_header_get_data_size(h);

	/* Get actual size chunk from file rather than from config */
	csz = cf_chunk_header_get_chunk_size(h);

	/* Update file size and fs size of necessary */
	if (new_size > old_size) {
		pthread_mutex_lock(&info_lock);
    		cf_info_set_ucomp(info, cf_info_get_ucomp(info) - old_size + new_size);
		pthread_mutex_unlock(&info_lock);

		cf_chunk_header_set_data_size(h, new_size);
	}

	/* Write the header to disk, this will update the 
	   timestamps (atime, mtime, ctime) on the underlying fs as well ;) */
	cf_chunk_write_header(f);

	/* No to be compressed by us */
	if (cf_chunk_header_get_coding(h) == CF_RAW) {
		res = pwrite(cf_chunk_file_get_fd(f), buf, size, sizeof(Cf_chunk_header_base)+offset);

		if (res == -1)
			res = -errno;

		return res;
	}


    	cf_chunk_file_wrlock(f);

    	/* Index of first chunk */
    	start = offset / csz;
    	/* Index of last chunk */
    	stop = (offset+size-1) / csz;

	written = 0;
	left = size;
	i = start;

    	/* Preload all the chunks */
    	cf_chunk_preload_chunks(f, start, stop);

	/* WARNING: the file must be marked dirty in the loop! 
		    Intermediate flushes (during a get_chunk()) 
		    removes the dirty flag
	*/

	while (left) {
		d = cf_chunk_get_chunk(f, i);

		/* Dirty */
		cf_chunk_file_set_dirty(f, 1);

		i_buffer = cf_chunk_data_get_data(d);
		i_size = cf_chunk_data_get_end(d);

		/* Beginning of chunk */
		c_offset = 0;

		/* First chunk offset within this chunk */
		if (i == start) {
			c_offset = offset - i * csz;
		} 


		/* Entire chunk  [<------->] */
		c_len = csz;

		/* First chunk   [___<---->] */
		if (i == start)
			c_len = csz - c_offset;

		/* Last chunk    [<---->___] */
		if (i == stop)
			c_len = left;


		/* Reduce amount left */
		left -= c_len;

		/* Mark chunk as modified */
		cf_chunk_data_set_dirty(d, 1);

		/* Set end of current chunk */
		if ((c_offset+c_len) > i_size)	
			cf_chunk_data_set_end(d, c_offset+c_len);

#ifdef CF_DEBUG
		cf_printf("start=%d stop=%d i=%d co=%d len=%d i_size=%d\n", start, stop, i, c_offset, c_len, i_size);
#endif
		/* copy */
		memcpy(i_buffer+c_offset, buf+written, c_len);

		written += c_len;
		i++;
	}

	/* Sanity check: size should be zero at this point */
	if (left) {
		cf_printf("ERROR: size = %d at end of %s\n", left, __FUNCTION__);
	}

	if (written != size) {
		cf_printf("ERROR: %d written asked %d in %s\n", written, size, __FUNCTION__);
		cf_printf("ERROR: offset %d size %d, chunk %d to %d\n", offset, size, start, stop);
	}


    cf_chunk_unload_chunks(f);
    cf_chunk_file_unlock(f);

    return written;
}

static int cf_statfs(const char *path,  struct statvfs  *stbuf)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = statvfs(xpath, stbuf);
    if(res == -1)
        return -errno;

    /* We do not lock the mapping but we can not be very wrong anyway */
/* df: used = blocks - free */
/*
*/
    pthread_mutex_lock(&info_lock);
    stbuf->f_bfree = stbuf->f_blocks - (cf_info_get_ucomp(info) / stbuf->f_bsize);
    pthread_mutex_unlock(&info_lock);
    stbuf->f_bavail = stbuf->f_bfree;

    return 0;
}


static int cf_flush(const char *path, struct fuse_file_info *fi)
{
    Cf_chunk_file pos = get_chunk_file(fi);
    Cf_chunk_header h = cf_chunk_file_get_header(pos);

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, path);
#endif 

    if (cf_chunk_header_get_coding(h) != CF_RAW) {
	cf_chunk_file_wrlock(pos);
    	cf_chunk_flush(pos);
	cf_chunk_file_unlock(pos);
    }

    return 0;
}

static int cf_release(const char *path, struct fuse_file_info *fi)
{
    int res = 0;
    int fd;
    int out;
    Cf_chunk_file pos;

    pthread_mutex_lock(&cf_map_get_lock(map));

#ifdef CF_DEBUG
    cf_printf("%s %s\n", __FUNCTION__, path);
#endif 
    pos = get_chunk_file(fi);

    out = cf_map_del(map, pos);

    if (out == 0) {
	/* Close file */
    	fd = cf_chunk_file_get_fd(pos);
    	close(fd);

    	cf_chunk_close(pos);

    	if(res == -1) {
		pthread_mutex_unlock(&cf_map_get_lock(map));
		return -errno;
	}
    }

    pthread_mutex_unlock(&cf_map_get_lock(map));

    return 0;
}

static int cf_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    int res;
    (void) path;

/*
    if (isdatasync)
        res = fdatasync(cf_chunk_file_get_fd(pos));
    else
        res = fsync(cf_chunk_file_get_fd(pos));
*/

    if(res == -1)
        return -errno;

    return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int cf_setxattr(const char *path, const char *name, const char *value,
                        size_t size, int flags)
{
    int res;

    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = lsetxattr(xpath, name, value, size, flags);
    if(res == -1)
        return -errno;
    return 0;
}

static int cf_getxattr(const char *path, const char *name, char *value,
                    size_t size)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = lgetxattr(path, name, value, size);
    if(res == -1)
        return -errno;
    return res;
}

static int cf_listxattr(const char *path, char *list, size_t size)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = llistxattr(path, list, size);
    if(res == -1)
        return -errno;
    return res;
}

static int cf_removexattr(const char *path, const char *name)
{
    int res;
    char xpath[CF_PATH_MAX];

    cf_path(config, path, xpath);

    res = lremovexattr(path, name);
    if(res == -1)
        return -errno;
    return 0;
}
#endif /* HAVE_SETXATTR */


void cf_message(void) {
	/* A little ego tripping... */
	printf("   +---------------------------------------------------------------------+\n");
	printf(" /---------------------------------------------------------------------/\n");
	printf("+---------------------------------------------------------------------+\n");
	printf("|\n");
	printf("| 	compFUSEd %s version %d \n",  CF_NAME, CF_VERSION);
	printf("|		by Johan Parent  \n");
	printf("|\n");
	printf("| Please send bug reports, suggestion to compFUSEd.contact@gmail.com\n");
	printf("| \n");
	printf("|\t - DISCLAIMER: read it! \n");
	printf("|\t - You run this program at your own risk! \n");
	printf("|\t - Treat your backups with respect :-P \n");
	printf("|\t - DO NOT store valuable data on this EXPERIMENTAL filesystem \n");
	printf("|\t - NEVER modify anything in the backend directory \n");
	printf("|\t   while the compFUSEd filesystem is mounted\n");
	printf("|\n");
	printf("| Feedback at above mentioned address is welcome\n");
	printf("+---------------------------------------------------------------------\n");

#ifdef CF_DEBUG
	printf(" WARNING: this in debug mode level 1, extra output will be generated (log or screen).\n");
#endif
#ifdef CF_DEBUG_L2
	printf(" WARNING: this in debug mode level 2, a lot of output will be generated (log or screen).\n");
#endif
#ifdef CF_DEBUG_L3
	printf(" WARNING: this in debug mode level 3, expensive checks are enabled!!!.\n");
#endif
	printf("\n");
}

static void* cf_init(void)
{
	/*
		We open then delete the cf_fs_info.bin file!!
	*/
	char xpath[CF_PATH_MAX];
	int no_info;

	cf_path(config, "/.cf_fs_info.bin", xpath);
	cf_printf("Filesystem info file %s\n", xpath);

	no_info = cf_info_read(xpath, info);

	cf_map_init(map, config);

	if (cf_config_get_forced_scan(config) || no_info) {
		cf_path(config, "", xpath);
		cf_printf("Scanning %s please wait...", xpath);
		cf_scan(xpath, info, 0);
		cf_printf("done!\n");
	}

	cf_info_print(info);

	pthread_mutex_init(&info_lock, NULL);

	cf_message();

	return NULL;
}

static void cf_destroy(void *ptr)
{
	/*
		All this is useless crap right now, I save nothing usefull.
	*/
	char xpath[CF_PATH_MAX];

	cf_path(config, "/.cf_fs_info.bin", xpath);

	cf_info_write(xpath, info);

	cf_log_close();

}

static struct fuse_operations cf_oper = {
    .getattr	= cf_getattr,
    .readlink	= cf_readlink,
    .opendir	= cf_opendir,
    .readdir	= cf_readdir,
    .releasedir	= cf_releasedir,
    .mknod	= cf_mknod,
    .mkdir	= cf_mkdir,
    .symlink	= cf_symlink,
    .unlink	= cf_unlink,
    .rmdir	= cf_rmdir,
    .rename	= cf_rename,
    .link	= cf_link,
    .chmod	= cf_chmod,
    .chown	= cf_chown,
    .truncate	= cf_truncate,
    .utime	= cf_utime,
    .open	= cf_open,
    .read	= cf_read,
    .write	= cf_write,
    .statfs	= cf_statfs,
    .release	= cf_release,
    .flush	= cf_flush,
    .fsync	= cf_fsync,
/*
    .init	= cf_init,
    .destroy	= cf_destroy,
*/
#ifdef HAVE_SETXATTR
    .setxattr	= cf_setxattr,
    .getxattr	= cf_getxattr,
    .listxattr	= cf_listxattr,
    .removexattr= cf_removexattr,
#endif
};

/* Find the mount point index from the args */
int cf_mountpoint(int argc, char **argv) {
	int i=1;

	while (i<argc) {
		if (strcmp(argv[i], "-o") == 0) {
			/* Skip next param */
			i+=2;
		} else if (strcmp(argv[i], "-f") == 0) {
			i++;
		} else if (strcmp(argv[i], "-d") == 0) {
			i++;
		} else if (strcmp(argv[i], "-s") == 0) {
			i++;
		} else if (strcmp(argv[i], "-h") == 0) {
			i++;
		} else {
			return i;
		}
	}

	return -1;
}

int main(int argc, char *argv[])
{
	int ret;
	int mountindex;

	mountindex = cf_mountpoint(argc, argv);

	if (mountindex < 0) {
		printf("Can't find mount point in cmd line params :( \n");

		return 1;
	}

	/* Set default logging to stdout */
	cf_log_open(NULL);

	config_file = cf_config_file_alloc();
	map = cf_map_alloc();

	/* Parse the different configuration files:
		- central config /usr/local/etc/compFUSEd.conf
		- local config $HOME/.compFUSEd
	*/
	cf_config_file_read(config_file);

	/* Find the configuration for this mount */
	config = cf_config_file_find(config_file, argv[mountindex]);

	if (!config) {
		printf("No configuration found for %s, bailing out!\n", argv[mountindex]);

		return 1;
	}

	cf_config_print(config);

	/* Check backend dir exists, otherwise bail out */
	cf_config_check_backend(config);

	cf_config_load_plugins(config);

	cf_info_init(info);

	cf_init();

	ret = fuse_main(argc, argv, &cf_oper);

	cf_destroy(NULL);

	return ret;
}
