#define	_XOPEN_SOURCE	500

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include "cf_chunk_file.h"
#include "cf_log.h"

/* How many percent must a chunk shrink before we compact the file */
#define	SHRINK_THRESHOLD	30

	/*
	
	INPUT:

		h = header : only set the toc_offset and toc_size
		b : buffer  with chunks
		t : table of content with up to data compressed chunk sizes 
		ot : original toc before any changes were made to the chunks


	FUNCTION:

		This function should write the data to disk. You can use any heuristic to handle this job. At least four
		things should be done by this function: 

			1) determine the offset values of the chunks
			2) write the data chunks data to disk
			3) determine the toc offset (toc is written at the end of the file behind the data chunks)
			4) write the header (at the beginning of the file)
	
	READ THESE NOTES!

		NOTE:	the size of the TOCs can differ
		NOTE: 	all the chunk are NOT necessarily in the buffer! Watch out for NULL values

		NOTE: 	some of the chunks may not have changed (ie not dirty)
			so _MAYBE_ you do not need to write them again to disk
			use cf_chunk_data_get_dirty() to see whether the chunk was changed


		NOTE: 	the chunk in the buffer will be deallocated automatically. Chunks allocated by
			you are YOUR responsibility!
	*/

/*

*/
void chunk_writer(int fd, int chunk_size, Cf_chunk_header h, Cf_chunk_buffer b, Cf_chunk_toc t, Cf_chunk_toc ot) {
	int tsize;
	int otsize;
	int i;
	int j;
	Cf_chunk_data d;
	Cf_chunk_data tmp;
	int total;		/* Current total compressed size */
	int data_offset = sizeof(Cf_chunk_header_base);
	int stop;
	int end;
	int old_size, new_size;
	int threshold = ( chunk_size * SHRINK_THRESHOLD ) / 100;

	tsize = cf_chunk_toc_get_size(t);
	otsize = cf_chunk_toc_get_size(ot);

        /* The first flush() on a brand new file has no old toc... we fake it */
        if (cf_chunk_toc_get_size(ot) == 0) {
		/* We don't loose the pointer since this is a copy */
		ot = t;
                otsize = tsize;

		/* NOTE:
			Doing this is not dangerous since the chunk should all be in the buffer.
			So we don't need to read anything from disk :)
		*/
        }

	/* Compare the 2 toc: new and old */
	stop = tsize;
	if (otsize<stop) {
		/* But we only use the 'old' part of the old toc */
		stop = otsize;
	}

	/* Compute total compressed size + fill in toc */
	/*
		We try to be a bit smarter here. Depending on the data a chunk can shrink
		because the data it contains compresses better. The writer_isimple plugin
		will compact the file. This can be at the expense of a lot of data being
		move to save only a few bytes per chunk.

		The threshold determines PER CHUNK whether or not we will compact the file.

		This is far from perfect (or very clever) but it is an easy to implement
		strategy.
	*/
	total = data_offset;
	for (i=0; i<tsize; i++) {
		cf_chunk_toc_set_entry_offset(t, i, total);

		new_size = abs(cf_chunk_toc_get_entry_size(t, i));

		if (i<cf_chunk_toc_get_size(ot))
			old_size = abs(cf_chunk_toc_get_entry_size(ot, i));
		else 
			old_size = 0;			

		if (new_size < old_size) {
			if ((old_size - new_size) < threshold) {
				total += old_size;
			} else {
				total += new_size;
			}
		} else {
			total += new_size;
		}
	}

	cf_chunk_header_set_toc_offset(h, total);

	/* Write the header at beginning of the file */
	pwrite(fd, h, sizeof(Cf_chunk_header_base), 0);

	/* Go over all the chunks in the buffer and write them to disk */
	for (i=0; i<stop; i++) {
		d = NULL;

		/* Is chunk 'i' in the buffer  */
		if (i<cf_chunk_buffer_get_buf_sz(b))
			d = cf_chunk_buffer_get_buffer(b,i);
		
		/* It might not be in the buffer but has it be moved */
		if (!d) {
			/* It the toc value differ we must load it to move */
			if (cf_chunk_toc_get_entry_offset(t,i) != cf_chunk_toc_get_entry_offset(ot,i)) {
				/* Alloc new data_chunk */
				d = cf_chunk_data_alloc(chunk_size);

				/* Save in the buffer */
				cf_chunk_buffer_add(b, i, d);

				/* Set the size */
				cf_chunk_data_set_end(d, cf_chunk_toc_get_entry_size(ot,i));

				/* Read from disk */
				cf_chunk_read_raw(fd, ot, i, d);
#if CF_DEBUG
	printf("H reading %d chk@%d size %d (*)\n", i, cf_chunk_toc_get_entry_offset(ot,i), cf_chunk_toc_get_entry_size(ot,i));
#endif
			}
		}

		/* If d == NULL at this point then nothing needs to be done :) */
		if (!d)
			continue;

		/* Problem we can not just write the chunk, we risk
		   overwritting data that has not changed but must be moved */
		end = cf_chunk_toc_get_entry_offset(t,i) + abs(cf_chunk_toc_get_entry_size(t,i));

		/* Read ahead to avoid overwriting data on disk */
		j = i + 1;

		/* See whether the end of 'd' is bigger than the offset of the next chunk */
		while ( (j<otsize) && (cf_chunk_toc_get_entry_offset(ot,j) < end)) {
#if CF_DEBUG_L2
	cf_chunk_buffer_print(b);
#endif
			tmp = NULL;

			/* Check if next shunk was also modified ie. present in buffer */
			if (j<cf_chunk_buffer_get_buf_sz(b)) {
				tmp = cf_chunk_buffer_get_buffer(b,j);
			} 

			/* Read from disk if not in buffer */
			if (!tmp) {
#if CF_DEBUG
	printf("H reading %d chk@%d size %d\n", j, cf_chunk_toc_get_entry_offset(ot,j), cf_chunk_toc_get_entry_size(ot,j));
#endif
				/* Alloc new data_chunk */
				tmp = cf_chunk_data_alloc(chunk_size);

				/* Save in the buffer */
				cf_chunk_buffer_add(b, j, tmp);

				/* Set the size */
				cf_chunk_data_set_end(tmp, cf_chunk_toc_get_entry_size(ot,j));

				/* Read from disk */
				cf_chunk_read_raw(fd, ot, j, tmp);
			}

			j++;
		}


		/* Perfect data is in buffer: no risk */

#if CF_DEBUG
	printf("H writing %d chk@%d size %d\n", i, cf_chunk_toc_get_entry_offset(t,i), cf_chunk_toc_get_entry_size(t,i));
#endif
		/* Write to disk */
		cf_chunk_write_raw(fd, t, i, d);

		/* Free mem */
		cf_chunk_data_dealloc(d);
		/* Remove from buffer entry */
		cf_chunk_buffer_del(b, i);

		/* Update old toc */
/*
		cf_chunk_toc_set_entry_offset(ot,i,cf_chunk_toc_get_entry_offset(t,i));
		cf_chunk_toc_set_entry_size(ot,i,cf_chunk_toc_get_entry_size(t,i));
*/
	} 


	/* If t > ot then the may be some chunks left */
	for (i=stop; i<tsize; i++) {
		/* Is chunk 'i' in the buffer  */
		if (i<cf_chunk_buffer_get_buf_sz(b))
			d = cf_chunk_buffer_get_buffer(b,i);
		
		if (d) {
#if CF_DEBUG
	printf("T writing %d chk@%d size %d\n", i, cf_chunk_toc_get_entry_offset(t,i), cf_chunk_toc_get_entry_size(t,i));
#endif
			/* Write to disk */
			cf_chunk_write_raw(fd, t, i, d);

			/* Free mem */
			cf_chunk_data_dealloc(d);
			/* Remove from buffer entry */
			cf_chunk_buffer_del(b, i);

			/* Update old toc */
/*
			cf_chunk_toc_set_entry_offset(ot,i,cf_chunk_toc_get_entry_offset(t,i));
			cf_chunk_toc_set_entry_size(ot,i,cf_chunk_toc_get_entry_size(t,i));
*/
		}
	}

	/* Last but not least write the toc behind the data chunks */
	/* Write the toc */
	pwrite(fd, cf_chunk_toc_get_entries(t), tsize*sizeof(Cf_toc_entry), total);


	/* Truncate (in case file got smaller */
	ftruncate(fd, (tsize*sizeof(Cf_toc_entry))+total);
}
